from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super
from future import standard_library
standard_library.install_aliases()
from builtins import object
from sqlite3 import OperationalError
from time import sleep, time

from rover import __version__
from .args import START, DAEMON, ROVERCMD, RECHECKPERIOD, PREINDEX, POSTSUMMARY, fail_early, STOP, UserFeedback, \
    FORCECMD
from .config import write_config
from .manager import DownloadManager
from .report import Reporter
from .index import Indexer
from .process import ProcessManager
from .sqlite import SqliteSupport
from .summary import Summarizer
from .utils import check_cmd, run, windows

"""
Commands related to the daemon:

The 'rover daemon' command - equivalent to related retrieval, but in the background.
The 'rover start' command - start the daemon,
The 'rover status' command - show whether the dameon is running.
The 'rover stop' command - stop teh daemon.
"""


DAEMONCONFIG = 'rover_daemon_config'
DOWNLOADCONFIG = 'rover_download_config'


class Starter(UserFeedback):
    """
### Start

Start the background (daemon) process to support `rover subscribe`.

See also `rover stop`, `rover status` and `rover daemon`.

##### Significant Parameters

@rover-cmd
@mseedindex-cmd
@download-workers
@mseedindex-workers
@temp-dir
@subscriptions-dir
@recheck-period
@download-retries
@http-timeout
@http-retries
@web
@http-bind-address
@http-port
@email
@email-from
@smtp-address
@smtp-port
@verbosity
@log-dir
@log-verbosity
@dev

In addition, parameters relevant to the processing pipeline (see `rover retrieve`, or the individual commands
for download, ingest and index) will apply,

Logging for individual processes in the pipeline will automatically configured with `--unique-logs --log-verbosity 3`.
For most worker tasks, that will give empty logs (no warnings or errors), which will be automatically deleted
(see `rover download`).  To preserve logs, and to use the provided verbosity level, start the daemon with `--dev`,

When the daemon is running status should be visible at http://localhost:8000 (by default).  When a subscription
is processed an email can be sent to the user (if `--email` is used).

#### Errors, Retries and Consistency

If `download-retries` allows, subscriptions are re-processed until no errors occur and, once data appear to be complete,
an additional retrieval is made which should result in no data being downloaded.  If this is not the case - if
additional data are found - then the web services are inconsistent.

Errors and inconsistencies are reported in the logs and in the optional email (`email` parameter) sent to the user.

##### Examples

    rover start -f roverrc

will start the daemon using the given configuration file.

    rover start --recheck-period 24

will start the daemon, processing subscriptions every 24 hours.

    """

    def __init__(self, config):
        UserFeedback.__init__(self, config)
        fail_early(config)
        self._rover_cmd = check_cmd(config, ROVERCMD, 'rover')
        # don't clean this because it may be long-lived (it will be over-written on re-use)
        self._config_path = write_config(config, DAEMONCONFIG, verbosity=0)

    def run(self, args):
        if args:
            raise Exception('Usage: rover %s' % START)
        if windows():
            run('%s %s -f %s' % ('pythonw -m rover', DAEMON, self._config_path), self._log, uncouple=True)
        else:
            run('%s %s -f %s' % (self._rover_cmd, DAEMON, self._config_path), self._log, uncouple=True)
        self._log.default('Rover version %s - starting %s' % (__version__, DAEMON))
        self.display_feedback()


class Stopper(object):
    """
### Stop

Stop the background (daemon) process to support `rover subscribe`.

See also `rover start`, `rover status`.

##### Significant Parameters

@verbosity
@log-dir
@log-verbosity

##### Examples

    rover stop -f roverrc

will stop the daemon that was started using the given configuration file.

    """

    def __init__(self, config):
        self._log = config.log
        self._processes = ProcessManager(config)

    def run(self, args):
        if args:
            raise Exception('Usage: rover %s' % STOP)
        self._processes.kill_daemon()  # logs suitable message


class StatusShower(object):
    """
### Status

Show whether the daemon is running or not..

See also `rover start`, `rover stop`.

##### Significant Parameters

@verbosity
@log-dir
@log-verbosity

##### Examples

    rover status -f roverrc

will show whether the daemon using the given configuration file is running.

    """

    def __init__(self, config):
        self._log = config.log
        self._processes = ProcessManager(config)

    def run(self, args):
        if args:
            raise Exception('Usage: rover %s' % STOP)
        print()
        print(self._processes.daemon_status())
        print()




class NoSubscription(Exception):
    """
    No suitable subscription found while scanning the database.
    """


class Daemon(SqliteSupport):
    """
### Daemon

The background (daemon) process that supports `rover subscribe`.

**Prefer using `rover start` to start this task in the background.**

See also `rover stop`, `rover status`.

##### Significant Parameters

@rover-cmd
@mseedindex-cmd
@download-workers
@mseedindex-workers
@temp-dir
@subscriptions-dir
@recheck-period
@download-retries
@http-timeout
@http-retries
@web
@http-bind-address
@http-port
@email
@email-from
@smtp-address
@smtp-port
@verbosity
@log-dir
@log-verbosity
@dev

In addition, parameters relevant to the processing pipeline (see `rover retrieve`, or the individual commands
for download, ingest and index) will apply,

Logging for individual processes in the pipeline will automatically configured with `--unique-logs --log-verbosity 3`.
For most worker tasks, that will give empty logs (no warnings or errors), which will be automatically deleted
(see `rover download`).  To preserve logs, and to use the provided verbosity level, start the daemon with `--dev`,

When the daemon is running status should be visible at http://localhost:8000 (by default).  When a subscription
is processed an email can be sent to the user (if `--email` is used).

#### Errors, Retries and Consistency

If `download-retries` allows, subscriptions are re-processed until no errors occur and, once data appear to be complete,
an additional retrieval is made which should result in no data being downloaded.  If this is not the case - if
additional data are found - then the web services are inconsistent.

Errors and inconsistencies are reported in the logs and in the optional email (`email` parameter) sent to the user.

##### Examples

    rover daemon -f roverrc

will start the daemon (in the foreground - see `rover start`) using the given configuration file.

    rover start --recheck-period 24

will start the daemon (in the foreground - see `rover start`), processing subscriptions every 24 hours.

    """

    def __init__(self, config):
        super().__init__(config)
        self._log = config.log
        self._pre_index = config.arg(PREINDEX)
        self._post_summary = config.arg(POSTSUMMARY)
        self._download_manager = DownloadManager(config, DOWNLOADCONFIG)
        self._recheck_period = config.arg(RECHECKPERIOD) * 60 * 60
        self._reporter = Reporter(config)
        self._config = config

    def run(self, args):
        if args:
            raise Exception('Usage: rover %s' % DAEMON)
        if self._pre_index:
            Indexer(self._config).run([])
        while True:
            try:
                try:
                    id = self._find_next_subscription()
                    self._add_subscription(id)
                except NoSubscription:
                    if self._download_manager.is_idle():
                        sleep(60)
                self._download_manager.step()
                sleep(1)
            except Exception as e:
                self._reporter.send_email('Rover Failure', self._reporter.describe_error(DAEMON, e))
                raise

    def _source_callback(self, source):
        self.execute('''update rover_subscriptions set last_error_count = ?, consistent = ? where id = ?''',
                     (source.errors.final_errors, source.consistent, source.name))
        if self._post_summary:
            Summarizer(self._config).run([])
        subject, msg = self._reporter.describe_daemon(source)
        self._reporter.send_email(subject, msg)

    def _find_next_subscription(self):
        now = time()
        found = [0]

        def callback(row):
            id = row[0]
            self._log.debug('Candidate: subscription %d' % id)
            if not found[0]:
                # todo - check this logic in test (try repeat while still downloading)
                if not self._download_manager.has_source(id):
                    found[0] = id

        try:
            self.foreachrow('''select id from rover_subscriptions
                                  where (last_check_epoch is NULL or last_check_epoch < ?)
                                  order by creation_epoch''', (now - self._recheck_period,),
                            callback, quiet=True)
        except OperationalError:
            pass  # no table exists, so no data

        if found[0]:
            return found[0]
        else:
            raise NoSubscription()

    def _add_subscription(self, id):
        try:
            path, availability_url, dataselect_url = self.fetchone(
                '''select file, availability_url, dataselect_url from rover_subscriptions where id = ?''', (id,))
            self._log.default('Adding subscription %d (%s, %s)' % (id, availability_url, dataselect_url))
            self._download_manager.add(id, path, True, availability_url, dataselect_url, self._source_callback)
        finally:
            self.execute('''update rover_subscriptions set last_check_epoch = ? where id = ?''', (time(), id))



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import super
from future import standard_library
standard_library.install_aliases()
from builtins import object
from os import O_WRONLY, open
from subprocess import Popen
from time import sleep

from .lock import DatabaseBasedLockFactory


"""
Support for running multiple sub-processes.
"""


class Workers(object):
    """
    A collection of processes that run asynchronously.  Note that the Python
    code here does NOT run asynchonously - it will block if there are no free
    workers.

    The idea is that (for example) we run N meedindex processes in the background
    so that we are not waiting on them to complete.
    """

    def __init__(self, config, n_workers):
        self._log = config.log
        self._n_workers = n_workers
        self._workers = []  # (command, popen, callback)

    def execute(self, command, callback=None):
        """
        Execute the command in a separate process.
        """
        self._wait_for_space()
        if not callback:
            callback = self._default_callback
        self._log.debug('Adding worker for "%s" (callback %s)' % (command, callback))
        self._workers.append((command, self._popen(command), callback))

    def _wait_for_space(self):
        while True:
            self.check()
            if self.has_space():
                self._log.debug('Space for new worker (%d/%d)' % (len(self._workers), self._n_workers))
                return
            sleep(0.1)

    def has_space(self):
        return len(self._workers) < self._n_workers

    def _default_callback(self, cmd, returncode):
        if returncode:
            raise Exception('"%s" returned %d' % (cmd, returncode))
        else:
            self._log.debug('"%s" succeeded' % (cmd,))

    def check(self):
        i = len(self._workers) - 1
        while i > -1:
            command, process, callback = self._workers[i]
            process.poll()
            if process.returncode is not None:
                self._log.debug('Calling callback %s (command %s)' % (callback, command))
                callback(command, process.returncode)
                self._workers = self._workers[:i] + self._workers[i+1:]
            i -= 1

    def wait_for_all(self):
        """
        Wait for all remaining processes to finish.
        """
        while True:
            self.check()
            if not self._workers:
                self._log.debug('No workers remain')
                return
            sleep(0.1)

    def _popen(self, command):
        return Popen(command, shell=True)


class NoConflictPerDatabaseWorkers(Workers):
    """
    Extend the above to block attempts for two processes (workers) to access
    the same resource.  Because this uses a database table it works across
    processes.
    """

    def __init__(self, config, n_workers, name):
        super().__init__(config, n_workers)
        self._lock_factory = DatabaseBasedLockFactory(config, name)
        self._locks = {}

    def execute_with_lock(self, command, key, callback=None):
        # here we are locking for subprocess, so we need to supply the PID
        # after acquiring the lock.
        # note that we cannot deadlock even though this is single-threaded and blocks,
        # because this is called only by code that iterates through the items to be locked
        # (so doesn't contain duplicates).  any contention must be with another process.
        self._wait_for_space()
        self._locks[key] = self._lock_factory.lock(key)
        self._locks[key].acquire()
        try:
            popen = self._popen(command)
            self._locks[key].set_pid(popen.pid)
        except:  # don't let a null PID stay in the table
            self._locks[key].release()
            raise
        self._log.debug('Adding worker for "%s" (callback %s)' % (command, callback))
        self._workers.append((command, popen,
                              lambda cmd, rtn: self._unlocking_callback(cmd, rtn, callback, key)))

    def execute(self, command, callback=None):
        raise Exception('Use execute_with_lock()')

    def _unlocking_callback(self, cmd, rtn, callback, key):
        self._log.debug('Unlocking %s' % key)
        # again, this is single thread - the locking is across processes
        self._locks[key].release()
        self._locks[key] = None
        if not callback:
            super()._default_callback(cmd, rtn)
        else:
            callback(cmd, rtn)




from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import open
from builtins import str
from future import standard_library
standard_library.install_aliases()
from builtins import object
from os.path import join, dirname
from sys import version_info

if version_info[0] >= 3:
    from tempfile import TemporaryDirectory
else:
    from backports.tempfile import TemporaryDirectory

from rover.config import BaseConfig, RepoInitializer
from rover.args import Arguments, TEMPDIR, DATADIR, FILE, INIT_REPOSITORY, DEFAULT_FILE
from rover.utils import canonify, windows

from .test_utils import WindowsTemp


class DummyLog(object):

    def debug(self, *args): pass
    def info(self, *args): pass
    def default(self, *args): pass
    def warn(self, *args): pass
    def error(self, *args): pass
    def critical(self, *args): pass


def test_write_config():
    with WindowsTemp(TemporaryDirectory) as dir:
        argparse = Arguments()
        args, config_path = argparse.parse_args([INIT_REPOSITORY])
        config = BaseConfig(DummyLog(), None, args, None, dir)
        RepoInitializer(config).run([dir])
        path = config.file(FILE)
        with open(path, 'r') as input:
            contents = input.read()
            assert contents == \
'''# specify configuration file
file=rover.config
# development mode (show exceptions)?
dev=False
# delete temporary files?
delete-files=True
# display help in markdown format?
md-format=False
# force cmd use (dangerous)
force-cmd=False
# the data directory - data, timeseries.sqlite
data-dir=data
# fractional increment for starting next timespan
timespan-inc=0.5
# fractional tolerance for overlapping timespans
timespan-tol=0.5
# maximum number of attempts to download data
download-retries=3
# number of download instances to run
download-workers=5
# command to run rover
rover-cmd=rover
# index before retrieval?
pre-index=True
# call ingest after retrieval?
ingest=True
# call index after ingest?
index=True
# call summary after retrieval?
post-summary=True
# availability service url
availability-url=http://service.iris.edu/irisws/availability/1/query
# dataselect service url
dataselect-url=http://service.iris.edu/fdsnws/dataselect/1/query
# temporary storage for downloads
temp-dir=tmp
# number of days before deleting temp files
temp-expire=1
# timeout for HTTP requests
http-timeout=60
# max retries for HTTP requests
http-retries=3
# force failures for testing (dangerous)
force-failures=0
# avoid OS sort (slower)?
sort-in-python=False
# process all files (not just modified)?
all=False
# when given a directory, process children?
recurse=True
# directory for subscriptions
subscriptions-dir=subscriptions
# time between availabilty checks
recheck-period=12
# skip overlap checks (dangerous)?
force-request=False
# directory for logs
log-dir=logs
# unique log names (with PIDs)?
log-unique=False
# number of days before deleting unique logs
log-unique-expire=7
# log verbosity (0-6)
log-verbosity=4
# maximum log size (e.g. 10M)
log-size=10M
# maximum number of logs
log-count=10
# console verbosity (0-6)
verbosity=4
# mseedindex command
mseedindex-cmd=mseedindex -sqlitebusyto 60000
# number of mseedindex instances to run
mseedindex-workers=10
# use leapseconds file?
leap=True
# number of days before refreshing file
leap-expire=30
# file for leapsecond data
leap-file=leap-seconds.list
# URL for leapsecond data
leap-url=http://www.ietf.org/timezones/data/leap-seconds.list
# auto-start the download progress web server?
web=True
# bind address for HTTP server
http-bind-address=127.0.0.1
# port for HTTP server
http-port=8000
# address for completion status
email=
# from address for email
email-from=noreply@rover
# address of SMTP server
smtp-address=localhost
# port for SMTP server
smtp-port=25
''', contents


def test_multiple_flags():
    argparse = Arguments()
    args, config_path = argparse.parse_args(['--index', '--no-index'])
    assert not args.index


def test_CONFIGDIR_start():
    with WindowsTemp(TemporaryDirectory) as dir:
        config = join(dir, DEFAULT_FILE)
        with open(config, 'w') as output:
            output.write('temp-dir=${CONFIGDIR}/foo\n')
            output.write('data-dir=$${CONFIGDIR}/foo\n')
        argparse = Arguments()
        args, config_path = argparse.parse_args(['-f', config])
        config = BaseConfig(None, None, args, None, dirname(config_path))
        assert config.dir(TEMPDIR)
        assert config.dir(TEMPDIR) == canonify(dir + '/foo'), config.dir(TEMPDIR)
        assert config.dir(DATADIR)
        assert config.dir(DATADIR) == canonify(join(dir, '${CONFIGDIR}/foo')), config.dir(DATADIR)


def test_CONFIGDIR_middle():
    if not windows():  # gets confused over messed up path
        with WindowsTemp(TemporaryDirectory) as dir:
            dir = canonify(dir)
            config = join(dir, '.rover')
            with open(config, 'w') as output:
                output.write('temp-dir=xx${CONFIGDIR}/foo\n')
                output.write('data-dir=xx$${CONFIGDIR}/foo\n')
            argparse = Arguments()
            args, config_path = argparse.parse_args(['-f', config])
            config = BaseConfig(None, None, args, None, dirname(config_path))
            assert config.dir(TEMPDIR)
            assert config.dir(TEMPDIR) == canonify(join(dir, 'xx' + dir + '/foo')), config.dir(TEMPDIR)
            assert config.dir(DATADIR)
            assert config.dir(DATADIR) == canonify(join(dir, 'xx${CONFIGDIR}/foo')), config.dir(DATADIR)


def test_CONFIGDIR_bad():
    with WindowsTemp(TemporaryDirectory) as dir:
        config = join(dir, '.rover')
        with open(config, 'w') as output:
            output.write('temp-dir=${FOO}\n')
        argparse = Arguments()
        args, config_path = argparse.parse_args(['-f', config])
        config = BaseConfig(None, None, args, None, dirname(config_path))
        try:
            config.arg(TEMPDIR)
        except Exception as e:
            assert 'does not exist' in str(e), str(e)
        else:
            assert False, 'Expected exception'



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import int
from future import standard_library
standard_library.install_aliases()
import sys
from re import match, sub
from sqlite3 import OperationalError

from .config import timeseries_db
from .args import MSEEDINDEXCMD, LEAP, LEAPEXPIRE, LEAPFILE, LEAPURL, DEV, VERBOSITY, MSEEDINDEXWORKERS, HTTPTIMEOUT, \
    HTTPRETRIES, FORCECMD, TIMESPANINC
from .args import TIMESPANTOL
from .coverage import MultipleSNCLBuilder
from .help import HelpFormatter
from .lock import MSEED
from .scan import ModifiedScanner, DirectoryScanner
from .sqlite import SqliteSupport
from .utils import format_epoch, windows, tidy_timestamp
from .utils import check_leap, check_cmd, STATION, NETWORK, CHANNEL, LOCATION
from .workers import NoConflictPerDatabaseWorkers


"""
Commands related to the index:

The 'rover index' command - call mseeedindex to update the tsindex table.
The 'rover list-index' command - displays entries from the tsindex table.
"""


class Indexer(ModifiedScanner, DirectoryScanner):
    """
### Index

    rover index [--all]

    rover index (file|dir)+

Index the files (add or change entires in the tsindex table in the mseed database).

When no argument is give all modified files in the repository are processed.  To force all files, use `--all`.

When a directory is given, all files contained in that directory are processed, along with the contents of
sub-directories, unless `--no-recurse` is specified.

The `mseedindex` command is used to index the data.  This optionally uses a file of leap-second data.  By default
(unless `--no-leap`) a file is downloaded from `--leap-url` if the file currently at `--leap-file` is missing or older
than `--leap-expire` days.

##### Significant Parameters

@all
@data-dir
@mseedindex-cmd
@mseedindex-workers
@leap
@leap-expire
@leap-file
@leap-url
@verbosity
@log-dir
@log-verbosity

##### Examples

    rover index --all

will index the entire repository.

"""

# Most of the work is done in the scanner superclasses which find the files
# to modify, and in the worker that runs mseedindex.

    def __init__(self, config):
        ModifiedScanner.__init__(self, config)
        DirectoryScanner.__init__(self, config)
        self._mseed_cmd = check_cmd(config, MSEEDINDEXCMD, 'mseedindex')
        self._timeseries_db = timeseries_db(config)
        self._leap_file = check_leap(config.arg(LEAP), config.arg(LEAPEXPIRE), config.arg(LEAPFILE),
                                     config.arg(LEAPURL), config.arg(HTTPTIMEOUT), config.arg(HTTPRETRIES), config.log)
        self._verbose = config.arg(DEV) and config.arg(VERBOSITY) == 5
        self._workers = NoConflictPerDatabaseWorkers(config, config.arg(MSEEDINDEXWORKERS), MSEED)

    def run(self, args):
        """
        Find the files to process (see superclasses).
        """
        if not args:
            self._log.info('Indexing all changed files')
            self.scan_data_dir()
        else:
            self.scan_dirs_and_files(args)

    def process(self, path):
        """
        Run mseedindex asynchronously in a worker.
        """
        self._log.info('Indexing %s' % path)
        if windows():
            self._workers.execute_with_lock('set LIBMSEED_LEAPSECOND_FILE=%s && %s %s -sqlite %s %s'
                                            % (self._leap_file, self._mseed_cmd, '-v -v' if self._verbose  else '',
                                               self._timeseries_db, path),
                                            path)
        else:
            self._workers.execute_with_lock('LIBMSEED_LEAPSECOND_FILE=%s %s %s -sqlite %s %s'
                                            % (self._leap_file, self._mseed_cmd, '-v -v' if self._verbose  else '',
                                               self._timeseries_db, path),
                                            path)

    def done(self):
        self._workers.wait_for_all()


BEGIN = 'begin'
END = 'end'
COUNT = 'count'
JOIN = 'join'
JOIN_QSR = 'join-qsr'
QUALITY = 'quality'
SAMPLERATE = 'samplerate'


class IndexLister(SqliteSupport, HelpFormatter):
    """
### List Index

    rover list-index [net=...|sta=...|loc=...|cha=..|qua=...|samp=...]* [begin=...] [end=...] \\
    [count|join|join-qsr]

    rover list-index [N_S_L_C_Q]* [begin=...] [end=...] \\
    [count|join|join-qsr]

List index entries for the repository (config parameter data-dir) that match the given constraints.

Note that console logging is to stderr, while the command results are listed to stdout.

#### Net_Sta_Chan_Loc_Qual and Samplerate

Query parameters can be named (network, station, location, channel, quality, samplerate) and unambiguous abbreviations
are accepted.  Alternatively, a N_S_L_C can be supplied (which can be truncated on the right, but must contain at least
one underscore).

The wildcards '*' and '?' can be used.

#### Time Range

The 'begin' and 'end' parameters can be given only once.  They must be of the form YYYY-MM-DDTHH:MM:SS.SSSSSS
(may be truncated on the right).  They define a range of times over which the data must appear (at least partially)
to be included:

#### Flags

The following parameters are simple flags that change the output format.  They are mutually exclusive and take no
value:

  count - only the number of matches will be shown

  join - continguous time ranges will be joined

  join-qsr - the maximal timespan across all quality and samplerates is shown (as used by retrieve)

##### Significant Parameters

@timespan-tol
@data-dir
@verbosity
@log-dir
@log-verbosity

#### Examples

    rover list-index IU_ANMO_00_BH? count

will display the number of entries for all time, and any quality or smaplerate.

    rover list-index net=* begin=2001-01-01

will list all entries in the index after the year 2000.

"""

    def __init__(self, config):
        SqliteSupport.__init__(self, config)
        HelpFormatter.__init__(self, False)
        self._timespan_inc = config.arg(TIMESPANINC)
        self._timespan_tol = config.arg(TIMESPANTOL)
        self._timeseries_db = timeseries_db(config)
        self._multiple_constraints = {STATION: [],
                                      NETWORK: [],
                                      CHANNEL: [],
                                      LOCATION: [],
                                      QUALITY: [],
                                      SAMPLERATE: []}
        self._single_constraints = {BEGIN: None,
                                    END: None}
        self._flags = {COUNT: False, JOIN: False, JOIN_QSR: False}

    def _display_help(self):
        self.print('''
The list-index command prints entries from the index that match 
the query parameters.  Parameters generally have the form 
name=value (no spaces).

The following parameters take '*' and '?' as wildcards, can be
repeated for multiple matches (combined with 'OR"), and the name
only has to match unambiguously (so cha=HHZ is OK):

  station, network, channel, location, quality, samplerate

The short form N_S_L_C_Q can also be used.

The following parameters can be given only once, must be of
the form YYYY-MM-DDTHH:MM:SS.SSSSSS (may be truncated on the
right), and define a range of times over which the block must 
appear (at least partially) to be included:

  begin, end

The following parameters are simple flags that change the
output format.  They are mutually exclusive and take no
value:

  count - only the number of matches will be shown

  join - contiguous time ranges will be joined

  join-qsr - the maximal timespan across all quality and 
  samplerates is shown (as used by retrieve) 

Examples:

    rover list-index IU_ANMO_00_BH? count

will display the number of entries for all time, any
quality or smaplerate.

    rover list-index net=* begin=2001-01-01

will list all entries in the index after the year 2000.
  
Note that console logging is to stderr, while results are
printed to stdout.

''')

    def run(self, args, stdout=None):
        if not stdout: stdout = sys.stdout
        if not args:
            self._display_help()
        else:
            self._check_database()
            self._parse_args(args)
            sql, params = self._build_query()
            if self._flags[COUNT]:
                self._count(sql, params, stdout=stdout)
            else:
                self._rows(sql, params, stdout=stdout)

    def _check_database(self):
        try:
            self.execute('select count(*) from tsindex')
        except OperationalError:
            raise Exception('''Cannot access the index table in the database (%s).  Bad configuration or no data indexed?''' % self._timeseries_db)

    def _parse_args(self, args):
        for arg in args:
            if arg in self._flags:
                self._assert_unset_flags(arg)
                self._flags[arg] = True
            else:
                parts = arg.split('=')
                if len(parts) == 1:
                    self._set_snclq(arg)
                elif len(parts) != 2:
                    raise Exception('Cannot parse "%s" (not of form name=value)' % arg)
                else:
                    name, value = parts
                    if name in (BEGIN, END):
                        self._set_time_limit(name, value)
                    else:
                        self._set_name_value(name, value)
        if self._flags[JOIN_QSR] and (
                self._multiple_constraints[SAMPLERATE] or self._multiple_constraints[QUALITY]):
            raise Exception('Cannot specify %s / %s AND %s' % (QUALITY, SAMPLERATE, JOIN_QSR))

    def _assert_unset_flags(self, name1):
        for name2 in self._flags:
            if self._flags[name2]:
                raise Exception('Cannot specify multiple keys (%s, %s)' % (name1, name2))

    def _set_snclq(self, snclq):
        components = snclq.split('_')
        if len(components) < 2 or len(components) > 5:
            raise Exception('Cannot parse %s (expect 2 to 5 values separated by "_" for N_S_L_C_Q)' % snclq)
        self._set_name_value(NETWORK, components[0])
        self._set_name_value(STATION, components[1])
        if len(components) > 2: self._set_name_value(LOCATION, components[2])
        if len(components) > 3: self._set_name_value(CHANNEL, components[3])
        if len(components) > 4: self._set_name_value(QUALITY, components[4])

    def _set_time_limit(self, name, value):
        if self._single_constraints[name]:
            raise Exception('Multiple values for %s' % name)
        value = tidy_timestamp(self._log, value)
        self._single_constraints[name] = value
        if self._single_constraints[BEGIN] and self._single_constraints[END] \
                and self._single_constraints[BEGIN] > self._single_constraints[END]:
            raise Exception('begin (%s) must be after end (%s)'
                            % (self._single_constraints[BEGIN], self._single_constraints[END]))

    def _set_name_value(self, name, value):
        if not match('^[\w\*\?]*$', value):
            raise Exception('Illegal characters in "%s"' % value)
        found = None
        for key in self._multiple_constraints.keys():
            if key.startswith(name):
                if found:
                    raise Exception('Ambiguous parameter: %s' % name)
                else:
                    found = key
        if not found:
            raise Exception('Did not recognise "%s"' % name)
        self._multiple_constraints[found].append(value)

    def _build_query(self):
        sql, params = 'select ', []
        if self._flags[COUNT]:
            sql += 'count(*) '
        else:
            sql += ('''network, station, location, channel, 
                       coalesce(timespans,'<' || starttime || ' ' || endtime || '>'), samplerate ''')
            if not self._flags[JOIN_QSR]:
                sql += ', quality '
        sql += 'from tsindex '
        constrained = False

        def conjunction(sql, constrained):
            if constrained:
                sql += 'and '
            else:
                sql += 'where '
                constrained = True
            return sql, constrained

        for name in self._multiple_constraints:
            repeated = False
            for value in self._multiple_constraints[name]:
                if not repeated:
                    sql, constrained = conjunction(sql, constrained)
                    sql += '( '
                    repeated = True
                else:
                    sql += 'or '
                sql += '%s like ? ' % name
                params.append(self._wildchars(value))
            if repeated:
                sql += ') '
        if self._single_constraints[BEGIN]:
            sql, constrained = conjunction(sql, constrained)
            sql += 'endtime > ?'
            params.append(self._single_constraints[BEGIN])
        if self._single_constraints[END]:
            sql, constrained = conjunction(sql, constrained)
            sql += 'starttime < ?'
            params.append(self._single_constraints[END])
        return sql, tuple(params)

    @staticmethod
    def _wildchars(value):
        return sub(r'\?', '_', sub(r'\*', '%', value))

    def _count(self, sql, params, stdout):
        # force to int here to avoid issues with strings on python 2
        print(int(self.fetchsingle(sql, params)), file=stdout)

    def _rows(self, sql, params, stdout):
        self._log.debug('%s %s' % (sql, params))
        builder = MultipleSNCLBuilder(self._log, self._timespan_tol, self._timespan_inc,
                                      self._flags[JOIN] or self._flags[JOIN_QSR])

        def callback(row):
            if self._flags[JOIN_QSR]:
                n, s, l, c, ts, r = row
                builder.add_timespans('%s_%s_%s_%s' % (n, s, l, c), ts, r)
            else:
                n, s, l, c, ts, r, q = row
                builder.add_timespans('%s_%s_%s_%s_%s (%g Hz)' % (n, s, l, c, q, r), ts, r)

        self.foreachrow(sql, params, callback)
        print()
        for coverage in builder.coverages():
            print('  %s' % coverage.sncl)
            for ts in coverage.timespans:
                print('    %s - %s' % (format_epoch(ts[0]), format_epoch(ts[1])))
            print()



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import open
from builtins import str
from future import standard_library
standard_library.install_aliases()
from datetime import datetime
from os import getpid, unlink
from os.path import exists, join
from re import match
from shutil import copyfile

from .args import MSEEDINDEXCMD, LEAP, LEAPEXPIRE, LEAPFILE, LEAPURL, DATADIR, INDEX, HTTPTIMEOUT, HTTPRETRIES
from .index import Indexer
from .lock import DatabaseBasedLockFactory, MSEED
from .scan import DirectoryScanner
from .sqlite import SqliteSupport, SqliteContext
from .utils import run, check_cmd, check_leap, create_parents, safe_unlink, windows, atomic_move

"""
The 'rover ingest' command - copy downloaded data into the repository (and then call index).
"""


# this is the table name when run directly from the command line.
# when run as a worker from (multiple) retriever(s) a table is supplied.
TMPFILE = 'rover_tmp_ingest'


class Ingester(SqliteSupport, DirectoryScanner):
    """
### Ingest

    rover ingest file

Add the contents of the file (MSEED format) to the repository and index the new data.

The `mseedindex` command is used to index the different blocks of dta present in the file.  THe corresponding byte
ranges are then appended to the appropriate files in the repository.

The file should not contain data that spans multiple calendar days.

##### Significant Parameters

@mseedindex-cmd
@data-dir
@index
@leap
@leap-expire
@leap-file
@leap-url
@verbosity
@log-dir
@log-verbosity

In addition, parameters for sub-commands (index) will be used - see help for those commands
for more details.

##### Examples

    rover ingest /tmp/IU.ANMO.00.*.mseed

will add all the data in the given file to the repository.

"""

# The simplest possible ingester:
# * Uses mseedindx to parse the file.
# * For each section, appends to any existing file using byte offsets
# * Refuses to handle blocks that cross day boundaries
# * Does not check for overlap, differences in sample rate, etc.

    def __init__(self, config):
        SqliteSupport.__init__(self, config)
        DirectoryScanner.__init__(self, config)
        self._mseed_cmd = check_cmd(config, MSEEDINDEXCMD, 'mseedindex')
        self._leap_file = check_leap(config.arg(LEAP), config.arg(LEAPEXPIRE), config.arg(LEAPFILE),
                                     config.arg(LEAPURL), config.arg(HTTPTIMEOUT), config.arg(HTTPRETRIES), config.log)
        self._db_path = None
        self._data_dir = config.dir(DATADIR)
        self._index = config.arg(INDEX)
        self._config = config
        self._log = config.log
        self._lock_factory = DatabaseBasedLockFactory(config, MSEED)

    def run(self, args, db_path=TMPFILE):
        """
        We only support explicit paths - modified file scanning makes no sense because
        the files are external, downloaded data.

        To avoid database contention we can run mseedindex into a unique database path.
        """
        self._db_path = db_path
        if not args:
            raise Exception('No paths provided')
        self.scan_dirs_and_files(args)

    def process(self, file):
        """
        Run mseedindex and, move across the bytes, and then call follow-up tasks.
        """
        self._log.info('Indexing %s for ingest' % file)
        if exists(self._db_path):
            self._log.warn('Temp file %s exists (deleting)' % self._db_path)
            safe_unlink(self._db_path)
        updated = set()
        try:
            if windows():
                run('set LIBMSEED_LEAPSECOND_FILE=%s && %s -sqlite %s %s'
                    % (self._leap_file, self._mseed_cmd, self._db_path, file), self._log)
            else:
                run('LIBMSEED_LEAPSECOND_FILE=%s %s -sqlite %s %s'
                    % (self._leap_file, self._mseed_cmd, self._db_path, file), self._log)
            with SqliteContext(self._db_path, self._log) as db:
                rows = db.fetchall('''select network, station, starttime, endtime, byteoffset, bytes
                                  from tsindex order by byteoffset''')
                updated.update(self._copy_all_rows(file, rows))
        finally:
            safe_unlink(self._db_path)
        if self._index:
            Indexer(self._config).run(updated)

    def _copy_all_rows(self, file, rows):
        self._log.info('Ingesting %s' % file)
        updated = set()
        with open(file, 'rb') as input:
            offset = 0
            for row in rows:
                offset, dest = self._copy_single_row(offset, input, file, *row)
                updated.add(dest)
        return updated

    def _copy_single_row(self, offset, input, file, network, station, starttime, endtime, byteoffset, bytes):
        self._assert_single_day(file, starttime, endtime)
        if offset < byteoffset:
            self._log.warn('Non-contiguous bytes in %s - skipping %d bytes' % (file, byteoffset - offset))
            input.read(byteoffset - offset)
            offset = byteoffset
        elif offset > byteoffset:
            raise Exception('Overlapping blocks in %s (mseedindex bug?)' % file)
        data = input.read(bytes)
        offset += bytes
        dest = self._make_destination(network, station, starttime)
        self._log.debug('Appending %d bytes from %s at offset %d to %s' % (bytes, file, byteoffset, dest))
        self._append_data(data, dest)
        return offset, dest

    def _make_destination(self, network, station, starttime):
        date_string = match(r'\d{4}-\d{2}-\d{2}', starttime).group(0)
        time_data = datetime.strptime(date_string, '%Y-%m-%d').timetuple()
        year, day = time_data.tm_year, time_data.tm_yday
        return join(self._data_dir, network, str(year), '%03d' % day, '%s.%s.%04d.%03d' % (station, network, year, day))

    def _append_data(self, data, dest):
        # here we are locking for this process, so we can set the PID directly.
        # there is no possibility for deadlock because we are single threaded
        # and release on exit.
        with self._lock_factory.lock(dest, pid=getpid()):
            # to avoid leaving broken files on unexpected exit, use a temp
            # file and then move into position (move should be atomic)
            tmp = dest + '.tmp'
            if exists(tmp):
                self._log.warn('Cleaning %s' % tmp)
                unlink(tmp)
            if not exists(dest):
                create_parents(dest)
                open(tmp, 'w').close()
            else:
                copyfile(dest, tmp)
            with open(tmp, 'ba') as output:
                output.write(data)
            atomic_move(self._log, tmp, dest)

    def _assert_single_day(self, file, starttime, endtime):
        if starttime[:10] != endtime[:10]:
            raise Exception('File %s contains data from more than one day (%s-%s)' % (file, starttime, endtime))



from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import

from builtins import super
from future import standard_library
standard_library.install_aliases()
from builtins import object
from re import sub

from .args import HELP, LIST_INDEX, DATADIR, INIT_REPOSITORY, RETRIEVE, TEMPDIR, INGEST, INDEX, SUBSCRIBE, \
    AVAILABILITYURL, DATASELECTURL, DOWNLOAD, LIST_RETRIEVE, mm, ALL, MSEEDINDEXCMD, Arguments, MDFORMAT, FILE, START, \
    STATUS, STOP, LIST_SUBSCRIBE, UNSUBSCRIBE, TRIGGER, DAEMON, LIST_SUMMARY, SUMMARY, DEFAULT_FILE, INIT_REPO, INIT

"""
The 'rover help' command.
"""


BACKGROUND = 'background'
USAGE = 'usage'
LOWLEVEL = 'low-level'


def welcome(config):
    return '''
                      Welcome to ROVER!

For more information on ROVER commands:
  
rover %s %s

  Displays information on the most common commands to immediately 
  download and ingest data.
    
rover %s %s

  Covers Rover's advanced mode, where it runs in the background, 
  continuously checking subscriptions and downloading data when 
  needed.
    
rover %s %s

  Lists lower-level commands that are less likely to be used.
  
Individual commands also have help.  See "rover %s %s".
    
For more information on configuration parameters, which can be 
provided via a file or using command line flags:

  rover -h

    Displays the command line parameters.
  
  %s

    Contains the defaults for these parameters and can be edited 
    to change the default behaviour.
    
To display this screen again, type "rover" or "rover help".

''' % (HELP, USAGE,
       HELP, BACKGROUND,
       HELP, LOWLEVEL,
       HELP, HELP,
       DEFAULT_FILE)


def usage(config):
    return '''
                    Common Rover Commands
                    
rover %s [directory]

  Initialize the given directory (or the current directory)
  as the repository.  This will create a configuration file
  (by default %s) as well as directories for logs and data.
  
  The aliases `rover %s` and `rover %s` also exist.

rover %s (file | sta=... [start [end]] | N_S_L_C [start [end]])

  Compare the local index with the data available remotely 
  (config parameter %s), then download (config parameter 
  %s) and ingest the missing files.  Use %s (below) to
  see what data would be downloaded (without doing the work).
  
rover %s (file | sta=... [start [end]] | N_S_L_C [start [end]])

  Compare the local index with the data available remotely 
  (config parameter %s), then display the difference.  
  Note that the summary is printed to stdout, while logging is 
  to stderr.

rover %s ...

  List index entries for the repository (config parameter 
  %s) that match the given constraints.  For more information, 
  run "rover %s" (with no arguments).

rover %s ...

  List summary entries for the repository (config parameter 
  %s) that match the given constraints.  This is faster than
  `rover %s` but gives less detail.  For more information, 
  run "rover %s" (with no arguments).

''' % (INIT_REPOSITORY, DEFAULT_FILE, INIT, INIT_REPO,
       RETRIEVE, AVAILABILITYURL, DATASELECTURL, LIST_RETRIEVE,
       LIST_RETRIEVE, AVAILABILITYURL,
       LIST_INDEX, DATADIR, LIST_INDEX,
       LIST_SUMMARY, DATADIR, LIST_INDEX, LIST_SUMMARY)


def background(config):
    return '''
                   Advanced Rover Commands
    
rover %s

  Start the background process that regularly downloads subscriptions.
  
rover %s

rover %s

  Display the status of, and stop, the background process.
                   
rover %s (file | sta=... [start [end]] | N_S_L_C [start [end]])

  Subscribe to retrieve updates whenever they become available.
  This has the same syntax as `rover %s`, but runs regularly in
  the background.
  
rover %s

rover %s N

  The first form lists the subscriptions.  The second shows
  which data will be downloaded for that subscription (similar to
  `rover %s`).
  
rover %s N

  Ask the background daemon to process the given subscription
  immediately, rather than waiting for the next update.
  
rover %s N

  Delete the subscription (the data remain in the repository, 
  but no more downloads will be made) 
  
''' % (START, STATUS, STOP,
       SUBSCRIBE, RETRIEVE,
       LIST_SUBSCRIBE, LIST_SUBSCRIBE, LIST_RETRIEVE,
       TRIGGER, UNSUBSCRIBE)


def low_level(config):
    return '''
                   Low-Level Rover Commands
                   
The following commands are used internally, but are usually not
useful from the command line:

rover %s url

  Download data from the given URL to the temporary store
  (config parameter %s).  When downloaded, ingest into the
  repository (config parameter %s) and delete.  Called
  by %s and the %s when needed.

rover %s (file|dir) ...

  Add the specified files to the repository (config
  parameter %s) and update the database index 
  Called by %s when needed.
  
rover %s [(file|dir) ...]

  Scan files and update the database index using the mseedindex 
  command (config parameter %s). Called by %s when needed.
  
  If no arguments are given then files in the repository
  (config parameter %s) that have been modified since the 
  repository was last indexed are processed.  The config 
  parameter %s can be used (eg %s on the command line) to force 
  processing of all files in the repository.
  
rover %s

  Re-generate the summary table used by `rover %s`.
      
''' % (DOWNLOAD, TEMPDIR, DATADIR, SUBSCRIBE, DAEMON,
       INGEST, DATADIR, RETRIEVE,
       INDEX, MSEEDINDEXCMD, INGEST, DATADIR, ALL, mm(ALL),
       SUMMARY, LIST_SUMMARY)


GENERAL = {
    USAGE: (usage, 'General interactive use'),
    BACKGROUND: (background, 'Advanced use with rover in the background'),
    LOWLEVEL: (low_level, 'Rarely used, low-level commands')
}


class HelpFormatter(object):
    """
    Either print markdown verbatim. or strip markdown and re-justify for 80 columns.
    """

    def __init__(self, md_format):
        self._md_format = md_format

    def print(self, text):
        arguments = Arguments()
        first_param = True
        for line in self.__paras(text):
            if line.startswith('@'):
                if first_param:
                    arguments.print_docs_header()
                    first_param = False
                if self._md_format:
                    arguments.print_docs_row_md(line[1:])
                else:
                    arguments.print_docs_row_text(line[1:])
            elif self._md_format:
                print(self.__escape(line))
            elif line.startswith('#'):
                print(line.lstrip(' #'))
            else:
                for short in self.__splitlines(line):
                    print(short)

    def __escape(self, text):
        text = sub(r'\\', '\\\\', text)
        # text = sub(r'`', '\\`', text)
        return text

    def __paras(self, text):
        lines = text.splitlines()
        i = 1
        while i < len(lines):
            if lines[i].strip() and lines[i-1].strip() and not lines[i].startswith('@') and not lines[i-1].endswith('\\'):
                lines[i-1] = lines[i-1].rstrip() + ' ' + lines[i].strip()
                lines[i:] = lines[i+1:]
            else:
                i += 1
        return lines

    def __slurp(self, line):
        word = ''
        space = ''
        while line and not line[0] == ' ':
            word += line[0]
            line = line[1:]
        while line and line[0] == ' ':
            space += line[0]
            line = line[1:]
        return word, space, line

    def __splitlines(self, line):
        indentation = ''
        while line and line.startswith(' '):
            indentation += line[0]
            line = line[1:]
        if not line:
            yield ''
            return
        while line:
            short, space = indentation, ''
            line = line.lstrip()
            while line:
                word, next_space, next_line = self.__slurp(line)
                if short.strip() and len(short + space + word) > 78:
                    yield short
                    short, space = indentation, ''
                else:
                    short = short + space + word
                    space, line = next_space, next_line
            if short:
                yield short


class Helper(HelpFormatter):

    def __init__(self, config):
        super().__init__(config.arg(MDFORMAT))
        self._config = config

    def run(self, args):
        from rover import COMMANDS   # avoid import loop
        if not args:
            self.print(welcome(self._config))
            return
        elif len(args) == 1:
            command = args[0].lower()
            if command == 'help':
                self._help()
                return
            if command in COMMANDS:
                self.print(COMMANDS[command][0].__doc__)
                return
            elif command in GENERAL:
                self.print(GENERAL[command][0](self._config))
                return
        raise Exception('Help is available for: %s, %s, %s, and individual commands (or simply "rover help")' % (USAGE, BACKGROUND, LOWLEVEL))

    def _help(self):
        from rover import COMMANDS   # avoid import loop
        self.print('''
### Help

Gives help on the various commands.
        
Help is available for the following commands:
        ''')
        for command in sorted(COMMANDS.keys()):
            print('%19s: %s' % (command, COMMANDS[command][1]))
        print('''
Help is also available for the following general topics: 
''')
        for command in GENERAL.keys():
            print('%19s: %s' % (command, GENERAL[command][1]))
        print('''
For example:

    rover help retrieve
''')



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import super
from future import standard_library
standard_library.install_aliases()
import datetime
from shutil import copyfile

from rover import __version__
from .args import RETRIEVE, TEMPDIR, AVAILABILITYURL, PREINDEX, LEAP, LEAPEXPIRE, UserFeedback, \
    LEAPFILE, LEAPURL, TEMPEXPIRE, LIST_RETRIEVE, DELETEFILES, POSTSUMMARY, DATASELECTURL, fail_early, HTTPTIMEOUT, \
    HTTPRETRIES
from .download import DEFAULT_NAME
from .index import Indexer
from .manager import DownloadManager, ManagerException
from .report import Reporter
from .sqlite import SqliteSupport
from .summary import Summarizer
from .utils import clean_old_files, match_prefixes, check_leap, unique_path, safe_unlink, build_file, fix_file_inplace

"""
Commands related to data retrieval:

The `rover retrieve` command - check for remote data that we don't already have, download it and ingest it.
The `rover list-retrieve` command - shows what data would be downloaded by `rover retrieve`.
"""


RETRIEVEWEB = 'rover_retrieve_availability'
RETRIEVECONFIG = 'rover_retrieve_config'
EARLY = datetime.datetime(1900, 1, 1)


class BaseRetriever(SqliteSupport, UserFeedback):
    """
### Retrieve

    rover retrieve file

    rover retrieve [net=N] [sta=S] [loc=L] [cha=C] [begin [end]]

    rover retrieve N_S_L_C [begin [end]]

Compare available data with the repository, then download, ingest and index data.

The file argument should contain a list of Net_Sta_Loc_Chans and timespans, as appropriate for calling an Availability
service (eg http://service.iris.edu/irisws/availability/1/).

In the second form above, at least one of `net`, `sta`, `loc`, `cha` should be given (missing values are
taken as wildcards).  For this and the third form a (single-line) file will be automatically constructed
containing that data.

The list of available data is retrieved from the service and compared with the local index.  Data not
available locally are downloaded and ingested.

In the comparison of available data, maximal timespans across all quality and sample rates are used (so quality
and samplerate information is "merged").

This command also indexes modified data in the repository before processing.

When the process is running status should be visible at http://localhost:8000 (by default).  When the process
ends an email can be sent to the user (if `--email` is used).

See `rover subscribe` for similar functionality, but with regular updates.

#### Errors, Retries and Consistency

If `download-retries` allows, retrievals are repeated until no errors occur and, once data appear to be complete,
an additional retrieval is made which should result in no data being downloaded.  If this is not the case - if
additional data are found - then the web services are inconsistent.

Errors and inconsistencies are reported in the logs and in the optional email (`email` parameter) sent to the user.
They also cause the command to exit with an error status.

##### Significant Parameters

@temp-dir
@availability-url
@dataselect-url
@timespan-tol
@pre-index
@ingest
@index
@post-summary
@rover-cmd
@mseedindex-cmd
@data-dir
@download-workers
@download-retries
@http-timeout
@http-retries
@leap-expire
@leap-file
@leap-url
@web
@http-bind-address
@http-port
@email
@email-from
@smtp-address
@smtp-port
@verbosity
@log-dir
@log-verbosity
@temp-expire

In addition, parameters for sub-commands (download, ingest, index) will be used - see help for those
commands for more details.

##### Examples

    rover retrieve N_S_L_C.txt

will download, ingest, and index any data missing from the repository for N_S_L_Cs / timespans present in the given file.

    rover retrieve IU_ANMO_00_BH1 2017-01-01 2017-01-04

will download, ingest and index and data for IU_ANMO_00_BH1 between the given dates that are missing from the repository.

"""

    def __init__(self, config):
        UserFeedback.__init__(self, config)
        SqliteSupport.__init__(self, config)
        fail_early(config)
        self._temp_dir = config.dir(TEMPDIR)
        self._availability_url = config.arg(AVAILABILITYURL)
        self._dataselect_url = config.arg(DATASELECTURL)
        self._pre_index = config.arg(PREINDEX)
        self._delete_files = config.arg(DELETEFILES)
        self._post_summary = config.arg(POSTSUMMARY)
        self._download_manager = None   # created in do_run()
        self._reporter = Reporter(config)
        self._config = config
        # leap seconds not used here, but avoids multiple threads all downloading later
        check_leap(config.arg(LEAP), config.arg(LEAPEXPIRE), config.file(LEAPFILE), config.arg(LEAPURL),
                   config.arg(HTTPTIMEOUT), config.arg(HTTPRETRIES), config.log)
        clean_old_files(self._temp_dir, config.arg(TEMPEXPIRE) * 60 * 60 * 24, match_prefixes(RETRIEVEWEB), config.log)

    def do_run(self, args, fetch, command):
        """
        Set-up environment, parse commands, and delegate to sub-methods as appropriate.
        """
        usage = 'Usage: rover %s (file | [net=N] [sta=S] [cha=C] [loc=L] [begin [end]] | N_S_L_C [begin [end]])' % command
        if not args:
            raise Exception(usage)
        # input is a temp file as we prepend parameters
        path = unique_path(self._temp_dir, RETRIEVEWEB, args[0])
        try:
            if len(args) == 1:
                copyfile(args[0], path)
            else:
                try:
                    build_file(self._log, path, args)
                except:
                    raise Exception(usage)
            fix_file_inplace(self._log, path, self._temp_dir)
            self._download_manager = DownloadManager(self._config, RETRIEVECONFIG if fetch else None)
            if fetch:
                self._log.default('Rover version %s - starting retrieve' % __version__)
                self.display_feedback()
            self._query(path, fetch)
            if fetch:
                return self._fetch()
            else:
                return self._display()
        except ManagerException:
            raise
        except Exception as e:
            self._reporter.send_email('Rover Failure', self._reporter.describe_error(RETRIEVE, e))
            raise
        finally:
            if self._delete_files:
                safe_unlink(path)

    def _query(self, up, fetch):
        """
        Populate the download manager by comparing the data from the
        availability service with the local index.
        """
        if self._pre_index:
            self._log.info('Ensuring index is current before retrieval')
            Indexer(self._config).run([])
        self._download_manager.add(DEFAULT_NAME, up, fetch,
                                   self._availability_url, self._dataselect_url, self._source_callback)

    def _fetch(self):
        """
        Fetch data from the download manager.
        """
        try:
            n_downloads = self._download_manager.download()
        finally:
            if self._post_summary:
                Summarizer(self._config).run([])
        return n_downloads

    def _display(self):
        """
        Display data from the download manager.
        """
        return self._download_manager.display()

    def _source_callback(self, source):
        subject, msg = self._reporter.describe_retrieve(source)
        self._reporter.send_email(subject, msg)


class Retriever(BaseRetriever):

    __doc__ = BaseRetriever.__doc__

    def __init__(self, config):
        super().__init__(config)

    def run(self, args):
        return self.do_run(args, True, RETRIEVE)


class ListRetriever(BaseRetriever):
    """
### List Retrieve

    rover list-retrieve file

    rover list-retrieve N_S_L_C [begin [end]]

Display what data would be downloaded if the `retrieve` equivalent command was run.

The file argument should contain a list of Net_Sta_Loc_Chans and timespans, as appropriate for calling an Availability
service (eg http://service.iris.edu/irisws/availability/1/).  Otherwise, if a N_S_L_C and timespan are given, a
(single-line) file will be automatically constructed containing that data.

##### Significant Parameters

@availability-url
@timespan-tol
@data-dir
@verbosity
@log-dir
@log-verbosity

##### Examples

    rover list-retrieve N_S_L_C.txt

will display the data missing form the repository to match what is available for the stations in the given file.

    rover list-retrieve IU.ANMO.00.BH1 2017-01-01 2017-01-04

will display the data missing from the repository to match what is available for IU.ANMO.00.BH1.

"""

    def __init__(self, config):
        super().__init__(config)

    def run(self, args):
        return self.do_run(args, False, LIST_RETRIEVE)




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

module_name = 'rover'

# Extract version string from code
import importlib
mod = importlib.import_module(module_name)
version = mod.__version__

setuptools.setup(
    name=module_name,
    version=version,
    author="IRIS",
    author_email="software-owner@iris.washington.edu",
    description="Tool for the robust retrieval of seismic data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://iris-edu.github.io/rover",
    packages=setuptools.find_packages(),
    classifiers=(
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ),
    entry_points={
        'console_scripts': [
            '%s = %s:main' % (module_name,module_name),
        ],
    },
    install_requires=[
        "requests",
        "nose",
        "future",
        "robotframework",
    ]
)



from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import dict
from builtins import str
from future import standard_library
standard_library.install_aliases()
pkg_path = __file__
__version__ = '0.0.7'

from traceback import print_exc

from .args import INIT_REPOSITORY, INDEX, INGEST, LIST_INDEX, \
    RETRIEVE, HELP, SUBSCRIBE, DOWNLOAD, LIST_RETRIEVE, START, STOP, LIST_SUBSCRIBE, UNSUBSCRIBE, DAEMON, \
    DEV, SUMMARY, LIST_SUMMARY, STATUS, WEB, TRIGGER, ABORT_CODE, ERROR_CODE
from .config import Config, RepoInitializer
from .daemon import Starter, Stopper, Daemon, StatusShower
from .download import Downloader
from .index import Indexer, IndexLister
from .ingest import Ingester
from .process import ProcessManager
from .retrieve import Retriever, ListRetriever
from .subscribe import Subscriber, SubscriptionLister, Unsubscriber, Trigger
from .summary import Summarizer, SummaryLister
from .web import ServerStarter


COMMANDS = {
    INIT_REPOSITORY: (RepoInitializer, 'Create / configure the repository'),
    INDEX: (Indexer, 'Index the repository'),
    INGEST: (Ingester, 'Ingest data from a file into the repository'),
    SUMMARY: (Summarizer, 'Update summary table'),
    LIST_INDEX: (IndexLister, 'List the contents of the repository'),
    LIST_SUMMARY: (SummaryLister, 'List a summary of the repository'),
    DOWNLOAD: (Downloader, 'Download data from a remote service'),
    RETRIEVE: (Retriever, 'Download, ingest and index missing data'),
    LIST_RETRIEVE: (ListRetriever, 'Show what data "rover retrieve" will download'),
    START: (Starter, 'Start the background daemon'),
    STOP: (Stopper, 'Stop the background daemon'),
    STATUS: (StatusShower, 'Show the background daemon status'),
    DAEMON: (Daemon, 'The background daemon (prefer start/stop)'),
    SUBSCRIBE: (Subscriber, 'Add a subscription'),
    LIST_SUBSCRIBE: (SubscriptionLister, 'List the subscriptions'),
    TRIGGER: (Trigger, 'Ask the daemon to reprocess subscriptions'),
    UNSUBSCRIBE: (Unsubscriber, 'Remove subscriptions'),
    WEB: (ServerStarter, 'Start a web server showing status')
}


def execute(command, config):
    from .help import Helper   # avoid import loop
    if not command:
        command = 'help'
    commands = dict(COMMANDS)
    commands[HELP] = (Helper, '')
    if command in commands:
        commands[command][0](config).run(config.args)
    else:
        raise Exception('Unknown command %s' % command)


def main():
    config = None
    try:
        try:
            config = Config()
            if config.command and config.command != HELP:
                config.lazy_validate()
            # initialise without a database...
            if not config.command or config.command in (INIT_REPOSITORY, HELP):
                execute(config.command, config)
            else:
                with ProcessManager(config):
                    execute(config.command, config)
        except KeyboardInterrupt:
            exit(ABORT_CODE)
    except Exception as e:
        if config and config.log:
            config.log.critical(str(e))
            if config.command in COMMANDS:
                config.log.default('See "rover help %s"' % config.command)
            elif config.command != HELP:
                config.log.default('See "rover help help" for a list of commands')
            if not config or not config._args or config.arg(DEV):
                print_exc()
        else:
            print_exc()
        # use an exit code so that workers are detected as failing
        exit(ERROR_CODE)
